==> sagua_0.2.3-1.tar.gz: sagua-0.2.3/src/player_mp3http.cc <==
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_mp3http.cc
/// @brief Class for play mp3 on HTTP source file.

#include <string>

#include <ace/Time_Value.h>

#include <libintl.h>

#include "../include/mainmanager.h"
#include "../include/player_mp3http.h"
#include "../include/scopeguard.h"
#include "../include/trace.h"

MainManager* MAD::MP3HttpPlayer::mainManager_ = NULL;

Glib::Mutex* MAD::MP3HttpPlayer::recvLoopMutex_ = NULL;
Glib::Cond* MAD::MP3HttpPlayer::recvLoopCondPush_ = NULL;
Glib::Cond* MAD::MP3HttpPlayer::recvLoopCondPop_ = NULL;

Glib::Dispatcher MAD::MP3HttpPlayer::signalUpdateStatusbar_;
Glib::Dispatcher MAD::MP3HttpPlayer::signalUpdateStatusbarTimeoutClear_;
Glib::Dispatcher MAD::MP3HttpPlayer::signalClearStatusbar_;

// struct mad_pcm has samples[2][1152], so is it enough 2304(1152*2) at
// initial size? 
std::vector<double> MAD::MP3HttpPlayer::bufSample_(2304);

DecodeState MAD::MP3HttpPlayer::decodeState_;
unsigned int MAD::MP3HttpPlayer::sampleRate_;
unsigned int MAD::MP3HttpPlayer::recvBufSize_;

bool MAD::MP3HttpPlayer::inRecving_;
bool MAD::MP3HttpPlayer::disconnected_;

char MAD::MP3HttpPlayer::recvBuf_[recvBufMaxSize_];
char MAD::MP3HttpPlayer::inputBuf_[inputBufSize_];

ACE_INET_Addr MAD::MP3HttpPlayer::destAddress_;
ACE_SOCK_Connector MAD::MP3HttpPlayer::connector_;
ACE_SOCK_Stream MAD::MP3HttpPlayer::peer_;

/// @brief Constructor of MP3HttpPlayer.
///
/// Reset static member variables and connect dispatchers to functions.
MAD::MP3HttpPlayer::MP3HttpPlayer(MainManager& mainManager)
    : recvLoopThread_(NULL)
{
    if(mainManager_ == NULL)
        mainManager_ = &mainManager;

    Reset();
    
    signalFinish_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayFinish));
    signalUpdateStatusbar_.connect(sigc::mem_fun(mainManager_, &MainManager::UpdateStatusbar));
    signalUpdateStatusbarTimeoutClear_.connect(sigc::mem_fun(mainManager_, &MainManager::UpdateStatusbarTimeoutClear));
    signalClearStatusbar_.connect(sigc::mem_fun(mainManager_, &MainManager::ClearStatusbar));
}

MAD::MP3HttpPlayer::~MP3HttpPlayer()
{
    if(decodeState_ == STOP)
        Clean();
    
}

/// @brief Initialize MP3HttpPlayer.
///
/// @param destUrl URL to open
/// @return Success of initialize
bool MAD::MP3HttpPlayer::Init(const std::string& destUrl)
{
    Tracer::Trace trace("MAD::MP3HttpPlayer::Init");

    NetUtil::HTTPHeaderStatus headerStatus = NetUtil::UNKNOWN_ERROR;
    unsigned int retry = 0;
    destUrl_ = destUrl;
    std::auto_ptr<Address> destAddressString;

    // clear statusbar at end of this function
    ScopeGuard timeoutClearStatusbar = MakeObjGuard(*mainManager_, &MainManager::TimeoutClearStatusbar);

    do {

        destAddressString = CreateAddress(destUrl_, mainManager_->ConvertUrlToCurrentEncoding());

        if(destAddressString.get() == NULL)
            return false;

        destAddress_.set(destAddressString->PortNumber(),
                         destAddressString->RawHostName());

        mainManager_->SetStatusbar(gettext("Connecting..."));

        if(connector_.connect(peer_, destAddress_) == -1) {

            mainManager_->SetStatusbar(gettext("Failed to connect."));

            return false;
        }

        if(destAddressString->Authorization() == NULL)
            NetUtil::SendGetMessage(destAddressString->Request(),
                                    peer_,
                                    "audio/mpeg, audio/x-mpegurl");
        else
            NetUtil::SendGetMessage(destAddressString->Request(),
                                    destAddressString->Authorization(),
                                    peer_,
                                    "audio/mpeg, audio/x-mpegurl");

        headerStatus = GetHeader();

    } while(headerStatus == NetUtil::REDIRECTION && retry++ < 3);

    if(headerStatus == NetUtil::SUCCESS)
        return true;
    else
        return false;
}

/// @brief Start player loop
/// 
/// Reset static member variables and start loop.
/// When play is finish, close audio and signal to MainManager.
void MAD::MP3HttpPlayer::Start()
{
    Tracer::Trace trace("MAD::MP3HttpPlayer::Start");

    // reset again cause this is static...
    Reset();
    decodeState_ = PLAY;

    mad_decoder_run(&decoder_, MAD_DECODER_MODE_SYNC);

    if(decodeState_ == PLAY)
        mainManager_->AudioClose(true);
    else
        mainManager_->AudioClose(false);
		
    Clean();
    signalFinish_();
}

/// @brief Set decodeState to STOP
///
/// Set decodeState to STOP. Header function detects STOP and send 
/// MAD_FLOW_STOP to decoder.
void MAD::MP3HttpPlayer::Stop()
{
    if(decodeState_ != STOP) {

    	decodeState_ = STOP;

        recvLoopCondPush_->broadcast();
    }
}

/// @brief Set decodeState to SEEK and set seekSeconds to seconds.
///
/// Do nothing while this is INET stream.
void MAD::MP3HttpPlayer::Seek(const unsigned long& seconds)
{}

/// @brief Clean up decoder.
void MAD::MP3HttpPlayer::Clean()
{
    mad_decoder_finish(&decoder_);

    peer_.close();

    buf_.data = NULL;
    buf_.length = 0;


    if(inRecving_) {
        inRecving_ = false;
        recvLoopCondPop_->broadcast();
        recvLoopCondPush_->broadcast();
        recvLoopThread_->join();
    }

    DeallocateThreadVariables();
}

/// @brief Reset static member variables.
///
/// Almost member variables are static, need to reset them by hands.
void MAD::MP3HttpPlayer::Reset()
{
    decodeState_ = STOP;

    sampleRate_ = 0;
}

/// @brief Recieve data from server and store to buffer.
/// @todo Adjust receive timeout.
void MAD::MP3HttpPlayer::RecvLoop()
{
    const ACE_Time_Value timeout(3, 0);
    ssize_t recvCnt = 0;
    disconnected_ = false;

    for(;;) {

        Glib::Mutex::Lock lock(*recvLoopMutex_);

        while(recvBufSize_ >= recvBufLimitSize_ && inRecving_)
            recvLoopCondPop_->wait(*recvLoopMutex_);

        if(!inRecving_)
            break;

        recvCnt = peer_.recv(recvBuf_ + recvBufSize_, 8192, &timeout);
        if(recvCnt <= 0) {

            mainManager_->SetStatusbarText(gettext("Disconnected."));
            signalUpdateStatusbarTimeoutClear_();

            disconnected_ = true;

            break;
        }

        recvBufSize_ += recvCnt;

        recvLoopCondPush_->signal();
    }

}

/// @brief Remove header.
/// @todo Get information from header.
/// @todo Consider end-of-line at buffer borderline.
NetUtil::HTTPHeaderStatus MAD::MP3HttpPlayer::GetHeader()
{
    const ACE_Time_Value timeout(10, 0);

    ssize_t recvCnt = peer_.recv(inputBuf_, sizeof(inputBuf_), &timeout);
    if(recvCnt <= 0) {

        mainManager_->SetStatusbar(gettext("Disconnected."));

        return NetUtil::UNKNOWN_ERROR;
    }

    char* separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(inputBuf_, ' ', recvCnt));
    if(separator == NULL)
        return NetUtil::UNKNOWN_ERROR;

    bool redirection = false;
    // All HTTP like protocol has status code at first.
    switch(separator[1]) {
    case '5':
        {
            char statusCode[4] = { 0, };
            ACE_OS::strncpy(statusCode, separator + 1, 3);
            Glib::ustring errStr(gettext("Server error : "));
            errStr += statusCode;

            mainManager_->SetStatusbar(errStr);
        }

        return NetUtil::SERVER_ERROR;
    case '4':
        {
            char statusCode[4] = { 0, };
            ACE_OS::strncpy(statusCode, separator + 1, 3);
            Glib::ustring errStr(gettext("Client error : "));
            errStr += statusCode;

            mainManager_->SetStatusbar(errStr);
        }

        return NetUtil::CLIENT_ERROR;
    case '3':
        redirection = true;

        break;
    case '2':

        break;
    case '1':
        // informational, not used

        return NetUtil::INFORMATIONAL;
    default:

        return NetUtil::UNKNOWN_ERROR;
    }

    for(;;) {

        separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(inputBuf_, '\r', recvCnt));

        if(redirection) {

            while(separator != NULL) {

                if(ACE_OS::memcmp(separator + 2, "Location: ", 10) == 0) {

                    char* endLine = ACE_reinterpret_cast(char*, ACE_OS::memchr(separator + 12, '\r', recvCnt - (separator - inputBuf_) - 12));
                    if(endLine != NULL) {

                        const size_t len = (endLine - separator) - 11;
                        char* location = new char[len];
                        ACE_OS::memset(location, 0, len);
                        ACE_OS::strncpy(location, separator + 12, len - 1);
                        destUrl_ = location;

                        // need close connection! or you can see hell, yeah.
                        peer_.close();

                        return NetUtil::REDIRECTION;
                    }
                }

                separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(separator + 1, '\r', recvCnt - (separator - inputBuf_) - 1)); 
            }

        } else {

            while(separator != NULL) {

                if(ACE_OS::memcmp(separator, "\r\n\r\n", 4) == 0) {

                    buf_.data = reinterpret_cast<unsigned char*>(separator + 4);
                    buf_.length = recvCnt - 4 - (separator - inputBuf_);

                    recvBufSize_ = 0;
                    inRecving_ = true;
                    AllocateThreadVariables();
                    recvLoopThread_ = Glib::Thread::create(sigc::mem_fun(*this, &MAD::MP3HttpPlayer::RecvLoop), true);

                    mad_decoder_init(&decoder_,
                                     &buf_,
                                     Input,
                                     Header,
                                     0,
                                     Output,
                                     Error,
                                     0);

                    return NetUtil::SUCCESS;

                }

                separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(separator + 1, '\r', recvCnt - (separator - inputBuf_) - 1));
            }

        }

        recvCnt = peer_.recv(inputBuf_, sizeof(inputBuf_), &timeout);

        if(recvCnt <= 0) {

            mainManager_->SetStatusbar(gettext("Disconnected."));

            return NetUtil::UNKNOWN_ERROR;
        }
    }
}

/// @brief Buffering file to decode.
///
/// inputBuf_ is main buffer to decode. recvBuf_ is receive buffer from
/// network in another thread. buf_.data is point to inputBuf_.
enum mad_flow MAD::MP3HttpPlayer::Input(void* data, struct mad_stream* stream)
{
    struct StreamBuffer* buf = static_cast<struct StreamBuffer*>(data);

    if(stream->next_frame) {

        buf->length = &buf->data[buf->length] - stream->next_frame;
        memmove(buf->data, stream->next_frame, buf->length);

    }

    const unsigned long cpySize = sizeof(inputBuf_) - buf->length;

    {
        Glib::Mutex::Lock lock(*recvLoopMutex_);

        if(recvBufSize_ < cpySize && !disconnected_) {

            mainManager_->SetStatusbarText(gettext("Buffering..."));
            signalUpdateStatusbar_();

            while(recvBufSize_ < recvBufLimitSize_ && decodeState_ != STOP && !disconnected_)
                recvLoopCondPush_->wait(*recvLoopMutex_);

            signalClearStatusbar_();
        }

        if(decodeState_ == STOP)
            return MAD_FLOW_BREAK;

        if(disconnected_) {

            if(recvBufSize_ > 0) {

                memcpy(&inputBuf_[buf->length],
                       recvBuf_,
                       recvBufSize_);

                buf->length = recvBufSize_;
                mad_stream_buffer(stream, buf->data, buf->length);

                recvBufSize_ = 0;

                return MAD_FLOW_CONTINUE;

            } else
                return MAD_FLOW_BREAK;
        }

        memcpy(&inputBuf_[buf->length],
               recvBuf_,
               cpySize);

        memmove(recvBuf_,
                &recvBuf_[cpySize],
                sizeof(recvBuf_) - cpySize);

        recvBufSize_ -= cpySize;

        recvLoopCondPop_->signal();
    }

    buf->length = sizeof(inputBuf_);

    mad_stream_buffer(stream, buf->data, buf->length);

    return MAD_FLOW_CONTINUE;
}

/// @brief Output decoded samples.
///
/// Check audio device and open device if need. Put samples to container and
/// send to play function. Update timer and signal to MainManager if timer
/// changed in second.
/// @todo Reduce duplicated codes. See MAD::MP3Player::Output
enum mad_flow MAD::MP3HttpPlayer::Output(void* data, struct mad_header const* header, struct mad_pcm* pcm)
{
    const unsigned int nchannels = pcm->channels;
    const unsigned int nsamples = pcm->length;
    const mad_fixed_t * left_ch = pcm->samples[0];
    const mad_fixed_t * right_ch = pcm->samples[1];

    if(!mainManager_->AudioCheckOpenDevice() || sampleRate_ != pcm->samplerate) {
        sampleRate_ = pcm->samplerate;
        mainManager_->AudioClose(true);
        if(!mainManager_->AudioOpenDevice(16, nchannels, pcm->samplerate, mainManager_->audioFormatNativeEndian_))
            return MAD_FLOW_STOP;
    }
    
    // ugly but for little more speed
    unsigned int index = 0;
    if(nchannels == 2) {
	
        for(unsigned int i = 0; i < nsamples; ++i) {
	    
            bufSample_[index] = *left_ch;
            bufSample_[index] += 0.5;
            bufSample_[index] /= 268435455.5;
            ++index;
            ++left_ch;

            bufSample_[index] = *right_ch;
            bufSample_[index] += 0.5;
            bufSample_[index] /= 268435455.5;
            ++index;
            ++right_ch;
        }
	
    } else {

        for(unsigned int i = 0; i < nsamples; ++i) {
	   	    
            bufSample_[index] = *left_ch;
            bufSample_[index] += 0.5;
            bufSample_[index] /= 268435455.5;
            ++index;
            ++left_ch;

            bufSample_[index] = 0.0;
            ++index;

        }

    }

    mainManager_->AudioPlay(bufSample_, nsamples);

    return MAD_FLOW_CONTINUE;
}

/// @brief Control decoder. 
enum mad_flow MAD::MP3HttpPlayer::Header(void* data, struct mad_header const* header)
{
    if(decodeState_ == STOP)
        return MAD_FLOW_STOP;
    
    return MAD_FLOW_CONTINUE;
}

/// @brief Check decoder's error.
enum mad_flow MAD::MP3HttpPlayer::Error(void* data, struct mad_stream* stream, struct mad_frame* frame)
{
    Tracer::Trace trace("MAD::MP3HttpPlayer::Error");
    
    if(stream->error == MAD_ERROR_BADCRC) {
        trace.Debug("MAD_ERROR_BAD_CRC");
		 
        return MAD_FLOW_IGNORE;

    }

    return MAD_FLOW_CONTINUE;
}

/// @brief Create MP3HttpPlayer object.
std::auto_ptr<Player> CreateMP3HttpPlayer(MainManager& mainManager)
{
    return std::auto_ptr<Player>(new MAD::MP3HttpPlayer(mainManager));
}

==> sagua_0.2.5-1.tar.gz: sagua-0.2.5/src/player_mp3http.cc <==
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_mp3http.cc
/// @brief Class for play mp3 on HTTP source file.

#include <string>

#include <ace/Time_Value.h>

#include <libintl.h>

#include "../include/mainmanager.h"
#include "../include/player_mp3http.h"
#include "../include/exception.h"
#include "../include/scopeguard.h"
#include "../include/trace.h"

template <class Buffer>
Buffer* MAD::MP3HttpPlayer<Buffer>::buffer_ = 0;

template <class Buffer>
PlayerInterface* MAD::MP3HttpPlayer<Buffer>::playerInterface_ = 0;
template <class Buffer>
MainManager* MAD::MP3HttpPlayer<Buffer>::mainManager_ = 0;

template <class Buffer>
Glib::Mutex* MAD::MP3HttpPlayer<Buffer>::recvLoopMutex_ = 0;
template <class Buffer>
Glib::Cond* MAD::MP3HttpPlayer<Buffer>::recvLoopCondPush_ = 0;
template <class Buffer>
Glib::Cond* MAD::MP3HttpPlayer<Buffer>::recvLoopCondPop_ = 0;

template <class Buffer>
Glib::Dispatcher MAD::MP3HttpPlayer<Buffer>::signalUpdateStatusbar_;
template <class Buffer>
Glib::Dispatcher MAD::MP3HttpPlayer<Buffer>::signalUpdateStatusbarTimeoutClear_;
template <class Buffer>
Glib::Dispatcher MAD::MP3HttpPlayer<Buffer>::signalClearStatusbar_;

template <class Buffer>
DecodeState MAD::MP3HttpPlayer<Buffer>::decodeState_;
template <class Buffer>
unsigned int MAD::MP3HttpPlayer<Buffer>::recvBufSize_;

template <class Buffer>
bool MAD::MP3HttpPlayer<Buffer>::inRecving_;
template <class Buffer>
bool MAD::MP3HttpPlayer<Buffer>::disconnected_;

template <class Buffer>
char MAD::MP3HttpPlayer<Buffer>::recvBuf_[recvBufMaxSize];
template <class Buffer>
char MAD::MP3HttpPlayer<Buffer>::inputBuf_[inputBufSize];

template <class Buffer>
ACE_INET_Addr MAD::MP3HttpPlayer<Buffer>::destAddress_;
template <class Buffer>
ACE_SOCK_Connector MAD::MP3HttpPlayer<Buffer>::connector_;
template <class Buffer>
ACE_SOCK_Stream MAD::MP3HttpPlayer<Buffer>::peer_;

/// @brief Constructor of MP3HttpPlayer.
///
/// Reset static member variables and connect dispatchers to functions.
template <class Buffer>
MAD::MP3HttpPlayer<Buffer>::MP3HttpPlayer(Buffer& buffer, const std::string& url)
    : recvLoopThread_(0),
      destUrl_(url)
{
    if(0 == buffer_)
        buffer_ = &buffer;

    if(0 == mainManager_) {

        mainManager_ = &MainManagerImpl::Instance();
        playerInterface_ = mainManager_;
    }

    try {

        Init();

    } catch(const Exception::NetException& ex) {

        peer_.close();

        throw;
    }

    signalFinish_.connect(sigc::mem_fun(playerInterface_, &PlayerInterface::Finished));
    signalUpdateStatusbar_.connect(sigc::mem_fun(mainManager_, &MainManager::UpdateStatusbar));
    signalUpdateStatusbarTimeoutClear_.connect(sigc::mem_fun(mainManager_, &MainManager::UpdateStatusbarTimeoutClear));
    signalClearStatusbar_.connect(sigc::mem_fun(mainManager_, &MainManager::ClearStatusbar));
}

/// @brief Initialize MP3HttpPlayer.
///
/// @param destUrl URL to open
/// @return Success of initialize
template <class Buffer>
void MAD::MP3HttpPlayer<Buffer>::Init()
{
    DEBUGTRACE("MAD::MP3HttpPlayer<Buffer>::Init");

    NetUtil::HTTPHeaderStatus headerStatus = NetUtil::UNKNOWN_ERROR;
    unsigned int retry = 0;
    std::auto_ptr<Address> destAddressString;

    ScopeGuard timeoutClearStatusbar = MakeObjGuard(*mainManager_, &MainManager::TimeoutClearStatusbar);

    do {

        destAddressString = CreateAddress(destUrl_, mainManager_->GetConvertUrlToCurrentEncoding());

        if(0 == destAddressString.get())
            throw Exception::NetException("Cannot create http address parser");

        destAddress_.set(destAddressString->GetPortNumber(),
                         destAddressString->GetRawHostName());

        mainManager_->SetStatusbar(gettext("Connecting..."));

        if(connector_.connect(peer_, destAddress_) == -1) {

            mainManager_->SetStatusbar(gettext("Failed to connect."));

            throw Exception::NetException("Cannot connect to address");
        }

        HttpRequestStrGenerator generator(*destAddressString, "audio/mpeg, audio/x-mpegurl");
        peer_.send(generator.GetString(), generator.GetLength());

        headerStatus = GetHeader();

    } while(headerStatus == NetUtil::REDIRECTION && retry++ < 3);

    if(headerStatus != NetUtil::SUCCESS)
        throw Exception::NetException("Cannot get success code");
}

/// @brief Start player loop
/// 
/// Reset static member variables and start loop.
/// When play is finish, close audio and signal to MainManager.
template <class Buffer>
void MAD::MP3HttpPlayer<Buffer>::DoStart()
{
    DEBUGTRACE("MAD::MP3HttpPlayer<Buffer>::Start");

    decodeState_ = PLAY;

    mad_decoder_run(&decoder_, MAD_DECODER_MODE_SYNC);

    Clean();

    buffer_->Close(decodeState_ == PLAY);
    
    signalFinish_();
}

/// @brief Set decodeState to STOP
///
/// Set decodeState to STOP. Header function detects STOP and send 
/// MAD_FLOW_STOP to decoder.
template <class Buffer>
void MAD::MP3HttpPlayer<Buffer>::DoStop()
{
    if(decodeState_ != STOP) {

    	decodeState_ = STOP;

        recvLoopCondPush_->broadcast();
    }
}

/// @brief Set decodeState to SEEK and set seekSeconds to seconds.
///
/// Do nothing while this is INET stream.
template <class Buffer>
void MAD::MP3HttpPlayer<Buffer>::DoSeek(unsigned long time)
{}

/// @brief Clean up decoder.
template <class Buffer>
void MAD::MP3HttpPlayer<Buffer>::Clean()
{
    mad_decoder_finish(&decoder_);

    peer_.close();

    buf_.data = 0;
    buf_.length = 0;

    if(inRecving_) {
        inRecving_ = false;
        recvLoopCondPop_->broadcast();
        recvLoopCondPush_->broadcast();
        recvLoopThread_->join();
    }

    DeallocateThreadVariables();
}

/// @brief Recieve data from server and store to buffer.
/// @todo Adjust receive timeout.
template <class Buffer>
void MAD::MP3HttpPlayer<Buffer>::RecvLoop()
{
    const ACE_Time_Value timeout(3, 0);
    ssize_t recvCnt = 0;
    disconnected_ = false;

    for(;;) {

        Glib::Mutex::Lock lock(*recvLoopMutex_);

        while(recvBufSize_ >= recvBufLimitSize && inRecving_)
            recvLoopCondPop_->wait(*recvLoopMutex_);

        if(!inRecving_)
            break;

        recvCnt = peer_.recv(recvBuf_ + recvBufSize_, 8192, &timeout);
        if(recvCnt <= 0) {

            mainManager_->SetStatusbarText(gettext("Disconnected."));
            signalUpdateStatusbarTimeoutClear_();

            disconnected_ = true;

            break;
        }

        recvBufSize_ += recvCnt;

        recvLoopCondPush_->signal();
    }

}

/// @brief Remove header.
/// @todo Get information from header.
/// @todo Consider end-of-line at buffer borderline.
template <class Buffer>
NetUtil::HTTPHeaderStatus MAD::MP3HttpPlayer<Buffer>::GetHeader()
{
    const ACE_Time_Value timeout(10, 0);

    ssize_t recvCnt = peer_.recv(inputBuf_, sizeof(inputBuf_), &timeout);
    if(recvCnt <= 0) {

        mainManager_->SetStatusbar(gettext("Disconnected."));

        return NetUtil::UNKNOWN_ERROR;
    }

    char* separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(inputBuf_, ' ', recvCnt));
    if(0 == separator)
        return NetUtil::UNKNOWN_ERROR;

    bool redirection = false;
    // All HTTP like protocol has status code at first.
    switch(separator[1]) {
    case '5':
        {
            char statusCode[4];
            ACE_OS::memcpy(statusCode, separator + 1, 3);
            statusCode[3] = '\0';

            Glib::ustring errStr(gettext("Server error : "));
            errStr += statusCode;

            mainManager_->SetStatusbar(errStr);
        }

        return NetUtil::SERVER_ERROR;
    case '4':
        {
            char statusCode[4];
            ACE_OS::memcpy(statusCode, separator + 1, 3);
            statusCode[3] = '\0';

            Glib::ustring errStr(gettext("Client error : "));
            errStr += statusCode;

            mainManager_->SetStatusbar(errStr);
        }

        return NetUtil::CLIENT_ERROR;
    case '3':
        redirection = true;

        break;
    case '2':

        break;
    case '1':
        // informational, not used

        return NetUtil::INFORMATIONAL;
    default:

        return NetUtil::UNKNOWN_ERROR;
    }

    for(;;) {

        separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(inputBuf_, '\r', recvCnt));

        if(redirection) {

            while(0 != separator) {

                if(ACE_OS::memcmp(separator + 2, "Location: ", 10) == 0) {

                    char* endLine = ACE_reinterpret_cast(char*, ACE_OS::memchr(separator + 12, '\r', recvCnt - (separator - inputBuf_) - 12));
                    if(0 != endLine) {

                        const size_t len = (endLine - separator) - 12;
                        char* location = new char[len + 1];
                        ACE_OS::memcpy(location, separator + 12, len);
                        location[len] = '\0';
                        destUrl_ = location;

                        // need close connection! or you can see hell, yeah.
                        peer_.close();

                        return NetUtil::REDIRECTION;
                    }
                }

                separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(separator + 1, '\r', recvCnt - (separator - inputBuf_) - 1)); 
            }

        } else {

            while(0 != separator) {

                if(ACE_OS::memcmp(separator, "\r\n\r\n", 4) == 0) {

                    buf_.data = reinterpret_cast<unsigned char*>(separator + 4);
                    buf_.length = recvCnt - 4 - (separator - inputBuf_);

                    recvBufSize_ = 0;
                    inRecving_ = true;
                    AllocateThreadVariables();
                    recvLoopThread_ = Glib::Thread::create(sigc::mem_fun(*this, &MAD::MP3HttpPlayer<Buffer>::RecvLoop), true);

                    mad_decoder_init(&decoder_,
                                     &buf_,
                                     Input,
                                     Header,
                                     0,
                                     Output,
                                     0,
                                     0);

                    return NetUtil::SUCCESS;

                }

                separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(separator + 1, '\r', recvCnt - (separator - inputBuf_) - 1));
            }

        }

        recvCnt = peer_.recv(inputBuf_, sizeof(inputBuf_), &timeout);

        if(recvCnt <= 0) {

            mainManager_->SetStatusbar(gettext("Disconnected."));

            return NetUtil::UNKNOWN_ERROR;
        }
    }
}

/// @brief Buffering file to decode.
///
/// inputBuf_ is main buffer to decode. recvBuf_ is receive buffer from
/// network in another thread. buf_.data is point to inputBuf_.
template <class Buffer>
enum mad_flow MAD::MP3HttpPlayer<Buffer>::Input(void* data, struct mad_stream* stream)
{
    struct StreamBuffer* buf = static_cast<struct StreamBuffer*>(data);

    if(stream->next_frame) {

        buf->length = &buf->data[buf->length] - stream->next_frame;
        memmove(buf->data, stream->next_frame, buf->length);

    }

    const unsigned long cpySize = sizeof(inputBuf_) - buf->length;

    {
        Glib::Mutex::Lock lock(*recvLoopMutex_);

        if(recvBufSize_ < cpySize && !disconnected_) {

            mainManager_->SetStatusbarText(gettext("Buffering..."));
            signalUpdateStatusbar_();

            while(recvBufSize_ < recvBufLimitSize && decodeState_ != STOP && !disconnected_)
                recvLoopCondPush_->wait(*recvLoopMutex_);

            signalClearStatusbar_();
        }

        if(decodeState_ == STOP)
            return MAD_FLOW_BREAK;

        if(disconnected_) {

            if(recvBufSize_ > 0) {

                memcpy(&inputBuf_[buf->length],
                       recvBuf_,
                       recvBufSize_);

                buf->length = recvBufSize_;
                mad_stream_buffer(stream, buf->data, buf->length);

                recvBufSize_ = 0;

                return MAD_FLOW_CONTINUE;

            } else
                return MAD_FLOW_BREAK;
        }

        memcpy(&inputBuf_[buf->length],
               recvBuf_,
               cpySize);

        memmove(recvBuf_,
                &recvBuf_[cpySize],
                sizeof(recvBuf_) - cpySize);

        recvBufSize_ -= cpySize;

        recvLoopCondPop_->signal();
    }

    buf->length = sizeof(inputBuf_);

    mad_stream_buffer(stream, buf->data, buf->length);

    return MAD_FLOW_CONTINUE;
}

/// @brief Output decoded samples.
///
/// Check audio device and open device if need. Put samples to container and
/// send to play function. Update timer and signal to MainManager if timer
/// changed in second.
/// @todo Reduce duplicated codes. See MAD::MP3Player::Output
template <class Buffer>
enum mad_flow MAD::MP3HttpPlayer<Buffer>::Output(void* data, struct mad_header const* header, struct mad_pcm* pcm)
{
    if(!playerInterface_->CheckOpenDevice()) {

        buffer_->Close(true);
        if(!playerInterface_->OpenDevice(16, pcm->channels, pcm->samplerate))
            return MAD_FLOW_STOP;
    }
    
    if(pcm->channels == 2) {

        buffer_->Push(pcm->samples[0], pcm->samples[1], pcm->length);	
	
    } else {

        buffer_->Push(pcm->samples[0], pcm->length);
    }

    return MAD_FLOW_CONTINUE;
}

/// @brief Control decoder. 
template <class Buffer>
enum mad_flow MAD::MP3HttpPlayer<Buffer>::Header(void* data, struct mad_header const* header)
{
    if(decodeState_ == STOP)
        return MAD_FLOW_STOP;
    
    return MAD_FLOW_CONTINUE;
}

/// @brief Create MP3HttpPlayer object.
template <class Buffer>
std::auto_ptr<Player> CreateMP3HttpPlayer(Buffer& buffer, const std::string& url)
{
    return std::auto_ptr<Player>(new MAD::MP3HttpPlayer<Buffer>(buffer, url));
}
